"""tweetstash

Usage:
  tweetstash [options] search
  tweetstash [options] listen

Options:
  --config=<config-dir>   Where to look for api.auth and hashtags.list.
  --stash=<stash-dir>     Where to save tweets.
  --by-user               Organize tweets by user id.
  --days=<days>           How far back to search.

"""
import sys
from docopt import docopt

from tweetstash import __version__, FileStash, TweetSearch


def main(argv=sys.argv):
    args = docopt(__doc__, version=__version__)

    stash = FileStash(
        base_dir=args['--stash'] or 'tweets',
        create_dir=True,
        by_user=args['--by-user'],
    )
    search = TweetSearch.from_config_dir(stash, args['--config'] or 'config')

    if args['search']:
        if args['--days']:
            stop_after = {'days': int(args['--days'])}
        else:
            stop_after = {}
        search.search(**stop_after)

    if args['listen']:
        search.listen()



from itertools import zip_longest
from abc import ABCMeta, abstractmethod
from pathlib import Path
import json


class Stash(metaclass=ABCMeta):
    """Abstract base class for tweetstash backends."""
    @abstractmethod
    def is_stashed(self, tweet_id):
        pass

    @abstractmethod
    def stash(self, tweet):
        pass

    def stash_many(self, tweets):
        for tweet in tweets:
            self.stash(tweet)

    @abstractmethod
    def unstash(self, tweet_id, **kwargs):
        pass

    def unstash_many(self, tweet_ids, kwargs_seq=()):
        for tweet_id, kwargs in zip_longest(tweet_ids, kwargs_seq):
            yield self.unstash(tweet_id, **kwargs or {})

    def unstash_all(self):
        yield from self.unstash_many(self.all_ids())

    @abstractmethod
    def remove_from_stash(self, tweet_id, **kwargs):
        pass


class FileStash(Stash):
    """"Filesystem tweetstash backend.
    Tweets are stored as separate JSON files, with the id as filename.

    """
    def __init__(self, base_dir=None, create_dir=False, by_user=True, log=True):
        self.by_user = by_user
        self.base_dir = Path('.') if base_dir is None else Path(base_dir)
        self.log = log
        if create_dir:
            self.base_dir.mkdir(parents=True, exist_ok=True)
        if not self.base_dir.exists():
            raise FileNotFoundError(base_dir)

        self.all_ids = set(self.read_all_ids())

    def tweet_path(self, tweet_id, user_id=None):
        path = self.base_dir
        if self.by_user:
            if user_id is None:
                raise TypeError('user_id required')
            path /= user_id

        return path / (tweet_id + '.json')

    def is_stashed(self, tweet_id, user_id=None):
        return tweet_id in self.all_ids

    def stash(self, tweet, overwrite=False):
        tweet_id = tweet['id_str']
        user_id = tweet['user']['id_str']
        if overwrite or not self.is_stashed(tweet_id, user_id=user_id):
            tweet_path = self.tweet_path(tweet_id, user_id=user_id)
            if self.by_user:
                tweet_path.parent.mkdir(exist_ok=True)
            with tweet_path.open('w') as file:
                json.dump(tweet, file)
            self.all_ids.add(tweet_id)
        if self.log:
            print('Saved tweet {}.'.format(tweet_id))

    def unstash(self, tweet_id, user_id=None):
        with self.tweet_path(tweet_id, user_id=user_id).open() as file:
            return json.load(file)

    def read_all_ids(self, user_id=None):
        if not self.by_user:
            paths = self.base_dir.glob('*.json')
        elif user_id is not None:
            paths = (self.base_dir / user_id).glob('*.json')
        else:
            paths = self.base_dir.glob('*/*.json')

        for tweet_path in paths:
            yield tweet_path.stem

    def remove_from_stash(self, tweet_id, user_id=None):
        self.tweet_path(tweet_id, user_id=user_id).unlink()
        self.all_ids.remove(tweet_id)



import sys
from pathlib import Path
from datetime import timedelta
from toolz import partition_all
import tweepy
from tqdm import tqdm

tweets_per_query = 100  # max allowed by Twitter API
max_hashtags_per_search = 30


class TweetSearch:
    def __init__(self, stash, auth_data, search_terms):
        self.stash = stash
        self.auth_data = auth_data
        self.search_terms = search_terms

    @classmethod
    def from_config_dir(cls, stash, config_dir=None):
        config_path = Path(config_dir or '.')

        # *.auth
        try:
            auth_path = next(config_path.glob('*.auth'))
        except StopIteration:
            raise FileNotFoundError('No .auth file found in {}'.format(config_path.absolute()))
        with auth_path.open() as auth_file:
            auth_data = auth_file.read().splitlines()

        # hashtags.list
        hashtags_path = config_path / 'hashtags.list'
        if not hashtags_path.is_file():
            raise FileNotFoundError(hashtags_path)
        with hashtags_path.open() as hashtags_file:
            hashtags = hashtags_file.read().splitlines()
        search_terms = ['#' + hashtag for hashtag in hashtags]

        return cls(stash, auth_data, search_terms)

    def search_api(self):
        auth = tweepy.AppAuthHandler(*self.auth_data[:2])
        api = tweepy.API(auth, wait_on_rate_limit=True, wait_on_rate_limit_notify=True)
        if not api:
            raise ValueError('Authentication failed')
        return api

    def stream_auth(self):
        auth = tweepy.OAuthHandler(*self.auth_data[:2])
        auth.set_access_token(*self.auth_data[2:])
        return auth

    def search(self, **stop_after):
        if not stop_after:
            stop_after['days'] = 36500
        stop_delta = timedelta(**stop_after)

        for hashtags in partition_all(max_hashtags_per_search, self.search_terms):
            tweets = search_twitter(self.search_api(), query=' OR '.join(hashtags))
            try:
                tweet = next(tweets)
            except StopIteration:
                break
            stop_time = tweet.created_at - stop_delta
            self.stash.stash(tweet._json)

            for tweet in tweets:
                if tweet.created_at < stop_time:
                    break
                self.stash.stash(tweet._json)

    def listen(self):
        listener = StashListener(self.stash)
        stream = tweepy.Stream(self.stream_auth(), listener)

        print('Listening (press Ctrl-C to stop)...')
        try:
            stream.filter(track=self.search_terms)
        except KeyboardInterrupt:
            sys.exit()


class StashListener(tweepy.streaming.StreamListener):
    def __init__(self, stash, *args, log=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.stash = stash
        self.log = log

    def on_status(self, tweet):
        self.stash.stash(tweet._json)
        if self.log:
            print('Saved tweet {}'.format(tweet.id_str))
        return True

    def on_error(self, status):
        print(status, file=sys.stderr)
        if status == 420:
            raise RuntimeError('Rate limited')


def search_twitter(api, max_results=None, progress=False, **query):
    """Perform a twitter search.

    See http://docs.tweepy.org/en/latest/api.html#API.search

    Parameters
    ----------
    api : tweepy.API
    max_results : int, optional
        The maximum number of tweets to return.
        Default is None (unlimited).
    progress : bool, optional
        Whether to show a progress bar.
    query : str, optional
        Query string.
    lang : str, optional
        Restrict tweets by language, given by an ISO 639-1 code.
    since_id : int, optional
        Return only statuses with ID greater than (i.e., more recent than) the specified ID.
    max_id : int, optional
        Return only statuses with ID less than (i.e., older than) the specified ID.
    geocode : str, optional
        Returns tweets by users located within a given radius.
        The parameter is specificed by 'latitude,longitude,radius' with radius followed by either 'km' or 'mi'.
        Note: see https://twittercommunity.com/t/search-api-returning-very-sparse-geocode-results/27998

    Yields
    ------
    tweepy.models.Status

    """
    query['q'] = query.pop('query')

    n_tweets_found = 0
    max_id = query.get('max_id', -1)
    if progress:
        progress_bar = tqdm(total=max_results, unit='tweets')
    while not max_results or n_tweets_found < max_results:
        try:
            if max_id > 0:
                query['max_id'] = str(max_id - 1)

            new_tweets = api.search(count=tweets_per_query, **query)
            if not new_tweets:
                break

            if progress:
                progress_bar.update(len(new_tweets))
            n_tweets_found += len(new_tweets)
            max_id = new_tweets[-1].id

            yield from new_tweets

        except KeyboardInterrupt:
            break

    if progress:
        progress_bar.close()
